@@ -21,8 +21,26 @@ except NameError:  # py3
     basestring = (bytes, str)
 
 logger = logging.getLogger(__name__)
+logger.setLevel(level=logging.WARNING)
 
 
+# https://metacpan.org/pod/AnyEvent::HTTP
+ANYEVENT_HTTP_STATUS_CODES = {
+    595: "Errors during connection establishment, proxy handshake",
+    596: "Errors during TLS negotiation, request sending and header processing",
+    597: "Errors during body receiving or processing",
+    598: "User aborted request via on_header or on_body",
+    599: "Other, usually nonretryable, errors (garbled URL etc.)"
+}
+
+SERVICES = {
+    "PVE": {"supported_backends": ["https", "openssh", "ssh_paramiko"], "supported_https_auths": ["password", "token"], "default_port": 8006, "token_separator": "=", "ssh_additional_options": "--output-format json"},
+    "PMG": {"supported_backends": ["https", "openssh", "ssh_paramiko"], "supported_https_auths": ["password"], "default_port": 8006},
+    "PBS": {"supported_backends": ["https"], "supported_https_auths": ["password", "token"], "default_port": 8007, "token_separator": ":"}}
+
+def config_failure(message, *args):
+    raise NotImplementedError(message.format(*args))
+
 class ProxmoxResourceBase(object):
 
     def __getattr__(self, item):
@@ -42,7 +60,15 @@ class ProxmoxResourceBase(object):
 
 
 class ResourceException(Exception):
-    pass
+    def __init__(self, status_code, status_message, content, errors=None):
+        self.status_code = status_code
+        self.status_message = status_message
+        self.content = content
+        self.errors = errors
+        if errors != None:
+            content += " - {0}".format(errors)
+        message = "{0} {1}: {2}".format(status_code, status_message, content).strip()
+        super(ResourceException, self).__init__(message)
 
 
 class ProxmoxResource(ProxmoxResourceBase):
@@ -63,7 +89,7 @@ class ProxmoxResource(ProxmoxResourceBase):
         if resource_id is not None:
             kwargs["base_url"] = self.url_join(self._store["base_url"], *resource_id)
 
-        return self.__class__(**kwargs)
+        return ProxmoxResource(**kwargs)
 
     def _request(self, method, data=None, params=None):
         url = self._store["base_url"]
@@ -75,8 +101,21 @@ class ProxmoxResource(ProxmoxResourceBase):
         logger.debug('Status code: %s, output: %s', resp.status_code, resp.content)
 
         if resp.status_code >= 400:
-            raise ResourceException("{0} {1}: {2}".format(resp.status_code, httplib.responses[resp.status_code],
-                                                          resp.content))
+            if hasattr(resp, 'reason'):
+                raise ResourceException(
+                    resp.status_code,
+                    httplib.responses.get(resp.status_code,
+                                        ANYEVENT_HTTP_STATUS_CODES.get(resp.status_code)),
+                    resp.reason,
+                    (self._store["serializer"].loads(resp) or {}).get('errors')
+                )
+            else:
+                raise ResourceException(
+                    resp.status_code,
+                    httplib.responses.get(resp.status_code,
+                                          ANYEVENT_HTTP_STATUS_CODES.get(resp.status_code)),
+                    resp.text
+                )
         elif 200 <= resp.status_code <= 299:
             return self._store["serializer"].loads(resp)
 
@@ -99,11 +138,20 @@ class ProxmoxResource(ProxmoxResourceBase):
         return self.put(*args, **data)
 
 
-class ProxmoxAPI(ProxmoxResourceBase):
-    def __init__(self, host, backend='https', **kwargs):
+class ProxmoxAPI(ProxmoxResource):
+    def __init__(self, host, backend='https', service='PVE', **kwargs):
+        service = service.upper()
+
+        # throw error for unsupported services
+        if not service in SERVICES.keys():
+            config_failure("{} service is not supported", service)
+
+        # throw error for unsupported backend for service
+        if not backend in SERVICES[service]["supported_backends"]:
+            config_failure("{} does not support {} backend", service, backend)
 
         #load backend module
-        self._backend = importlib.import_module('.backends.%s' % backend, 'proxmoxer').Backend(host, **kwargs)
+        self._backend = importlib.import_module('.backends.%s' % backend, 'proxmoxer').Backend(host, service=service, **kwargs)
         self._backend_name = backend
 
         self._store = {
@@ -115,7 +163,7 @@ class ProxmoxAPI(ProxmoxResourceBase):
     def get_tokens(self):
         """Return the auth and csrf tokens.
 
-        Returns (None, None) if the backend is not https.
+        Returns (None, None) if the backend is not https using password authentication.
         """
         if self._backend_name != 'https':
             return None, None
